@@ -59,6 +59,10 @@ class OntologyUriValidator(argparse.Action):
 def configure_arg_parser():
     """Configure command line parser."""
     parser = argparse.ArgumentParser(description='Ontology toolkit.')
+    parser.add_argument('-k', '--insecure', action="store_true",
+                        help="Allow insecure server connections when using SSL")
+    parser.add_argument('-v', '--version', action="store_true",
+                        help="Report onto-tool version and exit")
     subparsers = parser.add_subparsers(help='sub-command help', dest='command')
 
     update_parser = subparsers.add_parser('update',
@@ -186,6 +190,10 @@ def configure_arg_parser():
                                      " SHACL shapes and colors them green on the graph. This detection relies"
                                      " on the presence of sh:targetClass targeting, and can be confused by"
                                      " complex logical shapes or Advanced SHACL features such as SPARQL queries.")
+    graphic_parser.add_argument("--link-concentrator-threshold", type=int, default=3,
+                                help="When the number links originating from the same class that share a "
+                                     "single predicate exceed this threshold (default 10), use more compact "
+                                     "display. Setting the value to 0 disables this behavior.")
     sampling_limits = graphic_parser.add_argument_group(title='Sampling Limits')
     sampling_limits.add_argument("--instance-limit", type=int, default=500000,
                                  help="Specify a limit on how many triples to consider that use any one"
@@ -226,9 +234,13 @@ def configure_arg_parser():
                                     " using --exclude.")
     graphic_parser.add_argument('-v', '--version', help="Version to place in graphic",
                                 action="store")
-    graphic_parser.add_argument('-w', '--wee', action="store_true",
-                                help="a version of the graphic with only core"
-                                " information about ontology and imports")
+    graphic_parser.add_argument('-w', '--wee', nargs="*", default=None,
+                                help="For ontologies matching the patterns specified, only render "
+                                     "the name and import information. If no patterns are specified, applies "
+                                     "to all ontologies.")
+    graphic_parser.add_argument('--hide', nargs="*", default=[],
+                                help="When visualizing data, hide classes and properties matching the regex"
+                                     "patterns specified with this option.")
     graphic_parser.add_argument("--no-image", action="store_true",
                                 help="Do not generate PNG image, only .dot output.")
     graphic_parser.add_argument("-t", "--title", action="store",
