@@ -21,6 +21,8 @@ from rdflib.plugins.sparql import prepareQuery
 from SPARQLWrapper import TURTLE
 import pyshacl
 from pyparsing import ParseException
+
+import onto_tool
 from .command_line import configure_arg_parser
 from .ontograph import OntoGraf
 from .mdutils import Markdown2HTML
@@ -282,12 +284,16 @@ def generate_graphic(action, onto_files, endpoint, **kwargs):
 
     Keyword Parameters
     ------------------
-    wee : boolean
-        If True, generate a compact ontology graph.
+    wee : list(string)
+        If None, full details for all ontologies (--wee omitted from command line).
+        If empty list (-wee specified with no arguments), all ontologies are rendered compact
+        Otherwise, compact display for any ontologies matching the regular expressions in the list
     outpath : string
         Path of directory where graph will be output.
     version : string
         Version to be used in graphic title.
+    hide: list(string)
+        List of regular expressions to control which classes and/or properties will be hidden from the data graphic.
     include: list(string)
         List of ontology URIs to include for schema graph, or named graphs to consider for data graph.
     include_pattern: list(string)
@@ -300,6 +306,9 @@ def generate_graphic(action, onto_files, endpoint, **kwargs):
         named graphs are matched for exclusion from a data graph.
     show_shacl: boolean
         If True, attempt to detect SHACL shapes matching classes and properties.
+    concentrate_links: int
+        When the number links originating from the same class that share a single predicate exceed this threshold,
+        use more compact display. Setting the value to 0 disables this behavior.
 
     Returns
     -------
@@ -396,7 +405,7 @@ def __perform_export__(output, output_format, paths, context=None,
                        versioned=versioned_defined_by)
 
     serialized = g.serialize(format=output_format)
-    output.write(serialized.decode(output.encoding))
+    output.write(serialized)
 
 
 class VarDict(dict):
@@ -456,7 +465,7 @@ def __bundle_file_list(action, variables, ignore_target=False):
         if not ignore_target:
             tgt_dir = action['target'].format(**variables)
             if not isdir(tgt_dir):
-                os.mkdir(tgt_dir)
+                os.makedirs(tgt_dir)
         else:
             # There are times when
             tgt_dir = None
@@ -638,11 +647,64 @@ def __bundle_graph__(action, variables):
         os.mkdir(documentation)
     compact = action['compact'] if 'compact' in action else False
     og = OntoGraf([f['inputFile'] for f in __bundle_file_list(action, variables)],
-                  outpath=documentation, wee=compact, title=title, version=version)
+                  outpath=documentation, wee=[] if compact else None, title=title, version=version)
     og.gather_schema_info_from_files()
     og.create_schema_graf()
 
 
+def __bundle_local_sparql_each__(action, variables, output, queries):
+    for in_out in __bundle_file_list(action, variables):
+        g = Graph()
+        onto_file = in_out['inputFile']
+        parse_rdf(g, onto_file)
+        logging.debug("Input graph size for %s is %d", in_out['inputFile'], len(g))
+
+        updated = False
+        for query_file, query_text in queries:
+            logging.debug("Applying %s to %s", query_file, in_out['inputFile'])
+            parsed_update = None
+            parsed_query = None
+            try:
+                parsed_update = __parse_update_query__(query_text)
+            except ParseException:
+                # Not a update
+                parsed_query = prepareQuery(query_text)
+
+            if parsed_update:
+                g.update(parsed_update)
+                __transfer_query_prefixes__(g, parsed_update)
+                updated = True
+            else:
+                results = g.query(
+                    parsed_query,
+                    initNs={'xsd': XSD, 'owl': OWL, 'rdfs': RDFS, 'skos': SKOS})
+
+                if results.vars is not None:
+                    # SELECT Query
+                    select_output = __determine_output_file_name__(in_out['outputFile'],
+                                                                   queries, query_file, suffix='csv')
+                    with open(select_output, 'w') as csv_file:
+                        __serialize_select_results__(csv_file, results)
+
+                elif results.graph is not None:
+                    # CONSTRUCT Query
+                    __transfer_query_prefixes__(results.graph, parsed_query)
+                    rdf_format, suffix = __determine_format_and_suffix(action)
+                    construct_output = __determine_output_file_name__(in_out['outputFile'],
+                                                                      queries, query_file, suffix=suffix)
+                    results.graph.serialize(destination=construct_output, format=rdf_format, encoding='utf-8')
+                else:
+                    raise Exception('Unknown query type: ' + query_text)
+
+        if updated:
+            if 'format' in action:
+                rdf_format = 'pretty-xml' if action['format'] == 'xml' else action['format']
+            else:
+                rdf_format = 'turtle'
+            logging.debug("Saving updated RDF to %s", in_out['outputFile'])
+            g.serialize(destination=in_out['outputFile'], format=rdf_format, encoding='utf-8')
+
+
 def __bundle_local_sparql__(action, variables, output, queries):
     g = __build_graph_from_inputs__(action, variables)
     updated = False
@@ -688,8 +750,13 @@ def __bundle_local_sparql__(action, variables, output, queries):
 
 
 def __transfer_query_prefixes__(g, parsed_update):
-    for prefix, uri in parsed_update.prologue.namespace_manager.namespaces():
-        g.bind(prefix, uri)
+    if isinstance(parsed_update, list):
+        for update in parsed_update:
+            for prefix, uri in update.prologue.namespace_manager.namespaces():
+                g.bind(prefix, uri)
+    else:
+        for prefix, uri in parsed_update.prologue.namespace_manager.namespaces():
+            g.bind(prefix, uri)
 
 
 def __determine_format_and_suffix(action: dict) -> tuple:
@@ -747,7 +814,10 @@ def __bundle_sparql__(action, variables):
     if 'endpoint' in action:
         __bundle_endpoint_sparql__(action, variables, output, queries)
     else:
-        __bundle_local_sparql__(action, variables, output, queries)
+        if 'eachFile' in action and action['eachFile']:
+            __bundle_local_sparql_each__(action, variables, output, queries)
+        else:
+            __bundle_local_sparql__(action, variables, output, queries)
 
 
 def __serialize_select_results__(output, results):
@@ -921,7 +991,7 @@ def __process_construct_validation__(output, fail_on_warning, stop_on_fail, quer
             # Treat 'target' as directory.
             target_dir = output.format(**variables)
             if not isdir(target_dir):
-                os.mkdir(target_dir)
+                os.makedirs(target_dir)
             base, _ = splitext(basename(query_file))
             construct_output = join(target_dir, base + '.ttl')
         else:
@@ -1150,13 +1220,18 @@ def bundle_ontology(command_line_variables, bundle_path):
         elif action['action'] == 'mkdir':
             path = action['directory'].format(**substituted)
             if not isdir(path):
-                os.mkdir(path)
+                os.makedirs(path)
         elif action['action'] in BUNDLE_ACTIONS:
             BUNDLE_ACTIONS[action['action']](action, substituted)
         else:
             raise Exception('Unknown action ' + str(action))
 
 
+def __suppress_ssl_certificate_check():
+    import ssl
+    ssl._create_default_https_context = ssl._create_unverified_context
+
+
 def export_ontology(args, output_format):
     """Export one or more files as a single output.
 
@@ -1227,18 +1302,25 @@ def output_updated_ontology(args, g, onto_file, orig_format, output_format):
                     encoding='utf-8')
     else:
         serialized = g.serialize(format=output_format)
-        args.output.write(serialized.decode(args.output.encoding))
+        args.output.write(serialized)
 
 
 def main(arguments):
     """Do the thing."""
     args = configure_arg_parser().parse_args(args=arguments)
 
-    if args.debug:
+    if 'debug' in args and args.debug:
         logging.basicConfig(level=logging.DEBUG)
     else:
         logging.basicConfig(level=logging.INFO)
 
+    if args.version:
+        logging.info("onto-tool v%s", onto_tool.VERSION)
+        return
+
+    if args.insecure:
+        __suppress_ssl_certificate_check()
+
     if args.command == 'bundle':
         bundle_ontology(args.variables, args.bundle)
         return
@@ -1248,7 +1330,8 @@ def main(arguments):
                          limit=args.instance_limit, threshold=args.predicate_threshold,
                          single_graph=args.single_ontology_graphs,
                          wee=args.wee, outpath=args.output, version=args.version,
-                         no_image=args.no_image, title=args.title,
+                         no_image=args.no_image, title=args.title, hide=args.hide,
+                         concentrate_links=args.link_concentrator_threshold,
                          include=args.include, exclude=args.exclude,
                          include_pattern=args.include_pattern,
                          exclude_pattern=args.exclude_pattern,
