@@ -31,7 +31,8 @@ from .sparql_utils import create_endpoint, select_query
 
 class OntoGraf:
     def __init__(self, files, repo=None, **kwargs):
-        self.wee = kwargs.get('wee', False)
+        self.wee = kwargs.get('wee', None)
+        self.hide = kwargs.get('hide')
         title = kwargs.get('title')
         version = kwargs.get('version')
         if not version:
@@ -49,6 +50,7 @@ class OntoGraf:
         self.data = None
         self.no_image = kwargs.get('no_image', False)
         self.single_graph = kwargs.get('single_graph', False)
+        self.concentrate_links = kwargs.get('concentrate_links')
 
         outpath = kwargs.get('outpath', '.')
         self.outpath = outpath
@@ -62,6 +64,10 @@ class OntoGraf:
         self.limit = kwargs.get('limit', 500000)
         self.threshold = kwargs.get('threshold', 10)
         self.node_data = {}
+        self.class_names = {}
+        self.class_counts = defaultdict(int)
+        self.inheritance = []
+        self.super_color = "blue"
         self.arrow_color = "darkorange2"
         self.shacl_color = "darkgreen"
         self.arrowhead = "vee"
@@ -71,7 +77,7 @@ class OntoGraf:
         self.include_pattern = kwargs.get('include_pattern')
         self.exclude_pattern = kwargs.get('exclude_pattern')
 
-        self.superclasses = defaultdict(list)
+        self.superclasses = defaultdict(set)
 
         self.show_shacl = kwargs.get('show_shacl')
         self.shapes = defaultdict(list)
@@ -96,9 +102,10 @@ class OntoGraf:
 
     def graph_select_query(self, query):
         """Execute SPARQL SELECT query on local data, return results as generator."""
+        logging.debug(f"Local Query\n {query}")
         results = self.data.query(query)
         for result in results:
-            yield dict((str(k), str(v)) for k, v in zip(results.vars, result))
+            yield dict((str(k), str(v) if v is not None else None) for k, v in zip(results.vars, result))
 
     @staticmethod
     def strip_uri(uri):
@@ -154,7 +161,7 @@ class OntoGraf:
             prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
             prefix gist: <https://ontologies.semanticarts.com/gist/>
 
-            select ?ontology ?entity ?type where {
+            select DISTINCT ?ontology ?entity ?type where {
               graph ?g {
                 ?ontology a owl:Ontology .
                 {
@@ -178,7 +185,7 @@ class OntoGraf:
             prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
             prefix gist: <https://ontologies.semanticarts.com/gist/>
 
-            select ?ontology ?entity ?type where {
+            select DISTINCT ?ontology ?entity ?type where {
               ?ontology a owl:Ontology .
               {
                 ?ontology owl:imports ?entity .
@@ -210,19 +217,20 @@ class OntoGraf:
         for ontology, props in onto_data.items():
             self.node_data[ontology]['ontology'] = ontology
             self.node_data[ontology]['ontologyName'] = self.strip_uri(ontology)
-            for key in set(mapping.values()).union(set(['gist_thingsList'])):
+            for key in set(mapping.values()).union({'gist_thingsList'}):
                 if key != 'imports':
-                    self.node_data[ontology][key] = "\\l".join(sorted(props[key])) if key in props else ''
+                    self.node_data[ontology][key] = "\\l".join(sorted(set(props[key]))) if key in props else ''
                 else:
                     self.node_data[ontology][key] = props[key] if key in props else []
         return self.node_data
 
-    def create_schema_graf(self, data_dict=None, wee=None):
-        if data_dict is None:
-            data_dict = self.node_data
-        if wee is None:
-            wee = self.wee
-        if wee:
+    def create_schema_graf(self):
+        data_dict = self.node_data
+        wee = self.wee
+        # When 'wee' is not specified at all, it will be None
+        # When 'wee' is for all ontologies, it will be an empty list
+        # Otherwise, it will contain a list of ontology URI patterns
+        if wee is not None and not wee:
             self.graf = pydot.Dot(graph_type='digraph',
                                   label=self.title,
                                   labelloc='t',
@@ -255,7 +263,10 @@ class OntoGraf:
                 annotation_properties = file_data["annotation_propertiesList"]
                 gist_things = file_data["gist_thingsList"]
                 imports = file_data["imports"]
-                if wee:
+                render_compact = wee is not None and (
+                        not wee or any(re.search(pat, ontology) for pat in wee)
+                )
+                if render_compact:
                     node = pydot.Node(ontology_name)
                 else:
                     ontology_info = "{{{}\\l\\l{}|{}|{}|{}|{}|{}}}".format(
@@ -310,6 +321,14 @@ class OntoGraf:
             print()
         sys.stdout.flush()
 
+    def hidden(self, uri):
+        return any(re.search(pat, uri) for pat in self.hide)
+
+    def deepest_class(self, class_uris: str):
+        _, deepest = max((self.inheritance.index(cls_uri) if cls_uri in self.inheritance else -1, cls_uri)
+                         for cls_uri in class_uris.split(' '))
+        return deepest
+
     def gather_instance_info(self):
         predicate_query = """
         prefix owl: <http://www.w3.org/2002/07/owl#>
@@ -345,10 +364,17 @@ class OntoGraf:
                 self.data.parse(file_path, format=guess_format(file_path))
             all_predicates = list(self.graph_select_query(predicate_query))
 
+        hidden_predicates = set(predicate['predicate'] for predicate in all_predicates
+                                if self.hidden(predicate['predicate']))
+        logging.debug("Hiding predicates: %s", hidden_predicates)
+        all_predicates = [predicate for predicate in all_predicates if predicate['predicate'] not in hidden_predicates]
+
         if not all_predicates:
             logging.warning('No interesting predicates found in %s', self.repo or ' specified files')
             return
 
+        self.build_class_hierarchy()
+
         for count, predicate_row in enumerate(all_predicates):
             predicate = predicate_row['predicate']
             predicate_str = predicate_row['label'] if predicate_row.get('label') \
@@ -366,7 +392,7 @@ class OntoGraf:
                 predicate_usage = list(self.graph_select_query(query_text))
             logging.debug("%s items returned for %s", len(predicate_usage), predicate)
             for usage in predicate_usage:
-                if 'src' not in usage or int(usage.get('num', 0)) < self.threshold:
+                if 'src' not in usage or usage['src'] is None or int(usage.get('num', 0)) < self.threshold:
                     continue
                 self.record_predicate_usage(predicate, predicate_str, usage)
 
@@ -376,13 +402,19 @@ class OntoGraf:
             self.print_progress_bar(len(all_predicates), len(all_predicates),
                                     prefix='Processing predicates:', suffix='Complete', length=50)
 
-        # This is for future functionality
+        self.prune_for_inheritance()
+
+        if self.show_shacl:
+            self.add_shacl_coloring()
+
+    def build_class_hierarchy(self):
         inheritance_query = """
         prefix owl: <http://www.w3.org/2002/07/owl#>
         prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
         prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
+        prefix skos: <http://www.w3.org/2004/02/skos/core#>
 
-        select ?class ?parent where {
+        select distinct ?class ?c_label ?parent ?p_label where {
           {
               ?class rdfs:subClassOf+ ?parent .
           }
@@ -392,35 +424,112 @@ class OntoGraf:
             ?parent a owl:Class
           }
           filter (!isblank(?class) && !isblank(?parent))
+          OPTIONAL { ?class rdfs:label|skos:prefLabel ?c_label }
+          OPTIONAL { ?parent rdfs:label|skos:prefLabel ?p_label }
         }
-        """  # noqa: F841
-        # for inheritance_info in self.select_query(inheritance_query):
-        #     self.superclasses[inheritance_info['class']].append(inheritance_info['parent'])
-        # TODO Coalesce parent and child class references?
+        """
+        if self.repo:
+            parents = list(self.select_query(inheritance_query))
+        else:
+            parents = list(self.graph_select_query(inheritance_query))
 
-        if self.show_shacl:
-            self.add_shacl_coloring()
+        for inheritance_info in parents:
+            self.superclasses[inheritance_info['class']].add(inheritance_info['parent'])
+            self.class_names[inheritance_info['class']] = \
+                inheritance_info.get('c_label') or self.strip_uri(inheritance_info['class'])
+            self.class_names[inheritance_info['parent']] = \
+                inheritance_info.get('p_label') or self.strip_uri(inheritance_info['parent'])
+
+        # Determine evaluation order, root classes to leaves
+        remaining_classes = set(self.superclasses.keys())
+        root_classes = set(parent for cls, parents in self.superclasses.items()
+                           for parent in parents
+                           if parent not in remaining_classes)
+        eval_order = list(root_classes)
+        while remaining_classes:
+            next_set = set(cls for cls in remaining_classes
+                           if all(parent in eval_order for parent in self.superclasses[cls]))
+            # Make superclasses transitive
+            for cls in next_set:
+                parents = set(self.superclasses[cls])
+                for parent in parents:
+                    self.superclasses[cls].update(self.superclasses.get(parent, set()))
+            eval_order.extend(next_set)
+            remaining_classes.difference_update(next_set)
+
+        logging.debug('Inheritance evaluation order:\n%s',
+                      "\n".join(
+                          f"\t{self.strip_uri(cls)}: {list(self.strip_uri(sup) for sup in self.superclasses[cls])}"
+                          for cls in eval_order))
+        self.inheritance = eval_order
+
+        class_query = self.create_class_count_query(self.limit)
+        if self.repo:
+            class_counts = list(self.select_query(class_query))
+        else:
+            class_counts = list(self.graph_select_query(class_query))
+        for instance_info in class_counts:
+            self.class_counts[self.deepest_class(instance_info['src'])] += int(instance_info['num'])
+
+    def prune_for_inheritance(self):
+        for cls in reversed(self.inheritance):
+            if cls not in self.node_data or not self.superclasses[cls]:
+                continue
+            self.merge_with_parent(cls, 'links')
+            self.merge_with_parent(cls, 'data')
+
+    def merge_with_parent(self, cls, link_type):
+        removed = []
+        for link, count in self.node_data[cls][link_type].items():
+            # Locate parents with same link
+            parents_with_link = list(parent for parent in self.superclasses[cls]
+                                     if parent in self.node_data and link in self.node_data[parent][link_type])
+            if not parents_with_link:
+                continue
+            # Move to the most immediate parent, to create proper superclass chains
+            _, max_parent_with_link = max((self.inheritance.index(parent), parent) for parent in parents_with_link)
+            self.node_data[max_parent_with_link][link_type][link] += count
+            self.node_data[cls]['supers'].add(max_parent_with_link)
+            removed.append(link)
+        for r in removed:
+            del self.node_data[cls][link_type][r]
 
     def record_predicate_usage(self, predicate, predicate_str, usage):
-        if usage['src'] not in self.node_data:
+        src_uri = self.deepest_class(usage['src'])
+        tgt_uri = self.deepest_class(usage['tgt']) if 'tgt' in usage else None
+        if self.hidden(src_uri) or (tgt_uri is not None and self.hidden(tgt_uri)):
+            if tgt_uri is not None:
+                logging.debug('Hiding %s to %s link', src_uri, tgt_uri)
+            else:
+                logging.debug('Hiding %s attribute %s', src_uri, usage['dt'])
+            return
+        if src_uri not in self.node_data:
+            if src_uri is None:
+                raise "None src_uri in " + str(usage)
             src = {
-                'label': usage.get('srcLabel'),
-                'links': {},
-                'data': {}
+                'label': self.class_names.get(src_uri, self.strip_uri(src_uri)),
+                'links': defaultdict(int),
+                'data': defaultdict(int),
+                'supers': set()
             }
-            self.node_data[usage['src']] = src
+            self.node_data[src_uri] = src
         else:
-            src = self.node_data[usage['src']]
+            src = self.node_data[src_uri]
         if usage.get('dt'):
-            src['data'][(predicate, predicate_str, self.strip_uri(usage['dt']))] = int(usage['num'])
+            src['data'][(predicate,
+                         predicate_str or self.strip_uri(predicate),
+                         self.strip_uri(usage['dt']))] += int(usage['num'])
         else:
-            if usage['tgt'] not in self.node_data:
-                self.node_data[usage['tgt']] = {
-                    'label': usage.get('tgtLabel'),
-                    'links': {},
-                    'data': {}
+            if tgt_uri not in self.node_data:
+                if tgt_uri is None:
+                    raise "None src_uri in " + str(usage)
+                self.node_data[tgt_uri] = {
+                    'label': self.class_names.get(tgt_uri, self.strip_uri(tgt_uri)),
+                    'links': defaultdict(int),
+                    'data': defaultdict(int),
+                    'supers': set()
                 }
-            src['links'][(predicate, predicate_str, usage['tgt'])] = int(usage['num'])
+            src['links'][(predicate, predicate_str, tgt_uri)] += int(usage['num'])
 
     def add_shacl_coloring(self):
         shacl_query = """
@@ -437,6 +546,30 @@ class OntoGraf:
         for row in shacl_data:
             self.shapes[row['class']].append(row['property'])
 
+    def create_class_count_query(self, limit):
+        class_query = """
+            prefix owl: <http://www.w3.org/2002/07/owl#>
+            prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
+            prefix xsd: <http://www.w3.org/2001/XMLSchema#>
+            prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
+            prefix gist: <https://ontologies.semanticarts.com/gist/>
+            prefix skos: <http://www.w3.org/2004/02/skos/core#>
+
+            select ?src (COUNT(?src) as ?num) where {
+              {
+                select (group_concat(?o) as ?src) where {
+                  $pattern
+                  FILTER(!ISBLANK(?s))
+                  FILTER (!STRSTARTS(STR(?o), 'http://www.w3.org/2002/07/owl#'))
+                } group by ?s LIMIT $limit
+              }
+            } group by ?src
+            """
+        query_text = Template(class_query).substitute(
+            pattern=self.filtered_graph_pattern(str(RDF.type)),
+            limit=limit)
+        return query_text
+
     def create_predicate_query(self, predicate, predicate_type, limit):
         if predicate_type == str(OWL.ObjectProperty):
             type_query = """
@@ -447,26 +580,17 @@ class OntoGraf:
                 prefix gist: <https://ontologies.semanticarts.com/gist/>
                 prefix skos: <http://www.w3.org/2004/02/skos/core#>
 
-                select ?src ?srcLabel
-                       ?tgt ?tgtLabel
-                       ?num
-                where {
+                select ?src ?tgt (COUNT(?src) as ?num) where {
                   {
-                    select ?src ?tgt (COUNT(?src) as ?num) where {
-                      {
-                        select ?src ?tgt where {
-                          $pattern
-                          FILTER(!ISBLANK(?s))
-                          ?s a ?src .
-                          FILTER (!STRSTARTS(STR(?src), 'http://www.w3.org/2002/07/owl#'))
-                          ?o a ?tgt .
-                        } LIMIT $limit
-                      }
-                    } group by ?src ?tgt
+                    select (group_concat(?src_c) as ?src) (group_concat(?tgt_c) as ?tgt) where {
+                      $pattern
+                      FILTER(!ISBLANK(?s))
+                      ?s a ?src_c .
+                      FILTER (!STRSTARTS(STR(?src_c), 'http://www.w3.org/2002/07/owl#'))
+                      ?o a ?tgt_c .
+                    } group by ?s ?o LIMIT $limit
                   }
-                  OPTIONAL { ?src skos:prefLabel|rdfs:label ?srcLabel }
-                  OPTIONAL { ?tgt skos:prefLabel|rdfs:label ?tgtLabel }
-                }
+                } group by ?src ?tgt
                 """
         elif predicate_type == str(OWL.DatatypeProperty):
             type_query = """
@@ -477,25 +601,17 @@ class OntoGraf:
                 prefix gist: <https://ontologies.semanticarts.com/gist/>
                 prefix skos: <http://www.w3.org/2004/02/skos/core#>
 
-                select ?src ?srcLabel
-                       ?dt
-                       ?num
-                where {
+                select ?src ?dt (COUNT(?src) as ?num) where {
                   {
-                    select ?src ?dt (COUNT(?src) as ?num) where {
-                      {
-                        select ?src ?dt where {
-                          $pattern
-                          FILTER(!ISBLANK(?s))
-                          ?s a ?src .
-                          FILTER (!STRSTARTS(STR(?src), 'http://www.w3.org/2002/07/owl#'))
-                          BIND(DATATYPE(?o) as ?dt) .
-                        } LIMIT $limit
-                      }
-                    } group by ?src ?dt
+                    select (group_concat(?src_c) as ?src) (SAMPLE(?dtype) as ?dt) where {
+                      $pattern
+                      FILTER(!ISBLANK(?s) && ISLITERAL(?o))
+                      ?s a ?src_c .
+                      FILTER (!STRSTARTS(STR(?src_c), 'http://www.w3.org/2002/07/owl#'))
+                      BIND(DATATYPE(?o) as ?dtype) .
+                    } group by ?s LIMIT $limit
                   }
-                  OPTIONAL { ?src skos:prefLabel|rdfs:label ?srcLabel }
-                }
+                } group by ?src ?dt
                 """
         else:
             type_query = """
@@ -506,8 +622,8 @@ class OntoGraf:
                 prefix gist: <https://ontologies.semanticarts.com/gist/>
                 prefix skos: <http://www.w3.org/2004/02/skos/core#>
 
-                select ?src ?srcLabel
-                       ?tgt ?tgtLabel
+                select ?src
+                       ?tgt
                        ?dt
                        ?num
                 where {
@@ -515,36 +631,32 @@ class OntoGraf:
                     {
                       select ?src ?tgt (COUNT(?src) as ?num) where {
                         {
-                          select ?src ?tgt where {
-                            $pattern
-                            FILTER(!ISBLANK(?s))
-                            ?s a ?src .
-                            FILTER (!STRSTARTS(STR(?src), 'http://www.w3.org/2002/07/owl#'))
-                            ?o a ?tgt .
-                          } LIMIT $limit
+                            select (group_concat(?src_c) as ?src) (group_concat(?tgt_c) as ?tgt) where {
+                              $pattern
+                              FILTER(!ISBLANK(?s))
+                              ?s a ?src_c .
+                              FILTER (!STRSTARTS(STR(?src_c), 'http://www.w3.org/2002/07/owl#'))
+                              ?o a ?tgt_c .
+                            } group by ?s ?o LIMIT $limit
                         }
                       } group by ?src ?tgt
                     }
-                    OPTIONAL { ?src skos:prefLabel|rdfs:label ?srcLabel }
-                    OPTIONAL { ?tgt skos:prefLabel|rdfs:label ?tgtLabel }
                   }
                   UNION
                   {
                     {
                       select ?src ?dt (COUNT(?src) as ?num) where {
                         {
-                          select ?src ?dt where {
-                            $pattern
-                            FILTER(!ISBLANK(?s))
-                            FILTER(isLITERAL(?o))
-                            ?s a ?src .
-                            FILTER (!STRSTARTS(STR(?src), 'http://www.w3.org/2002/07/owl#'))
-                            BIND(DATATYPE(?o) as ?dt) .
-                          } LIMIT $limit
+                            select (group_concat(?src_c) as ?src) (SAMPLE(?dtype) as ?dt) where {
+                              $pattern
+                              FILTER(!ISBLANK(?s) && ISLITERAL(?o))
+                              ?s a ?src_c .
+                              FILTER (!STRSTARTS(STR(?src_c), 'http://www.w3.org/2002/07/owl#'))
+                              BIND(DATATYPE(?o) as ?dtype) .
+                            } group by ?s LIMIT $limit
                         }
                       } group by ?src ?dt
                     }
-                    OPTIONAL { ?src skos:prefLabel|rdfs:label ?srcLabel }
                   }
                 }
                 """
@@ -553,6 +665,7 @@ class OntoGraf:
             limit=limit)
         return query_text
 
+    MIN_LINE_WIDTH = 1
     MAX_LINE_WIDTH = 5
 
     @staticmethod
@@ -560,7 +673,30 @@ class OntoGraf:
         """Scale line width relative to the most commonly occurring edge for the graph"""
         if graph_max == 1:
             return OntoGraf.MAX_LINE_WIDTH
-        return min(5, max(1, round(log(num_used, pow(graph_max, 1.0/OntoGraf.MAX_LINE_WIDTH)))))
+        if num_used == 0:
+            return OntoGraf.MIN_LINE_WIDTH
+        try:
+            return min(OntoGraf.MAX_LINE_WIDTH,
+                       max(OntoGraf.MIN_LINE_WIDTH,
+                           round(log(num_used, pow(graph_max, 1.0 / OntoGraf.MAX_LINE_WIDTH)))))
+        except ValueError:
+            logging.warning('Failed to determine line width from num=%d max=%d', num_used, graph_max)
+            return OntoGraf.MIN_LINE_WIDTH
+
+    MAX_FONT_SIZE = 24
+    MIN_FONT_SIZE = 14
+
+    @staticmethod
+    def font_size(num_instances, graph_max):
+        """Scale line width relative to the most commonly occurring edge for the graph"""
+        if num_instances == 0 or graph_max == 1:
+            return OntoGraf.MIN_FONT_SIZE
+        span = OntoGraf.MAX_FONT_SIZE - OntoGraf.MIN_FONT_SIZE
+        try:
+            return OntoGraf.MIN_FONT_SIZE - 1 + min(span, max(1, round(log(num_instances, pow(graph_max, 1.0 / span)))))
+        except ValueError:
+            logging.warning('Failed to determine font size from num=%d max=%d', num_instances, graph_max)
+            return OntoGraf.MIN_FONT_SIZE
 
     def create_instance_graf(self, data_dict=None):
         self.graf = pydot.Dot(graph_type='digraph',
@@ -581,26 +717,36 @@ class OntoGraf:
         logging.debug("Node data: %s", data_dict)
         logging.debug("Shape data: %s", self.shapes)
 
+        if not data_dict:
+            logging.warning("No data found with the specified parameters.")
+            return
+
         # Determine the maximum number any edge occurs in the data, so the edge widths can be properly scaled
         max_common = max(occurs for class_data in data_dict.values() for occurs in class_data['links'].values())
 
+        max_instance = max(self.class_counts.values())
+
         for class_, class_data in data_dict.items():
+            node_font_size = self.font_size(self.class_counts[class_], max_instance)
+            node_line_width = self.line_width(self.class_counts[class_], max_instance)
             if class_data['data']:
                 class_info = \
-                    """<<table border="0" cellspacing="0" cellborder="1">
+                    """<<table color="black" border="{line_width}" cellspacing="0" cellborder="1">
                      <tr>
-                      <td align="center" bgcolor="{label_bg}"><font color="{label_fg}">{class_label}</font></td>
-                     </tr>
+                      <td align="center" bgcolor="{label_bg}"><font point-size="{fontsize}" color="{label_fg}">{class_label}</font></td></tr>
                      <tr>
                       <td align="center">{attribute_text}</td>
                      </tr>
                     </table>>""".format(
                         label_fg="white" if class_ in self.shapes else "black",
                         label_bg="darkgreen" if class_ in self.shapes else "white",
+                        fontsize=node_font_size,
+                        line_width=node_line_width,
                         class_label=class_data['label'] if class_data['label'] else self.strip_uri(class_),
                         attribute_text="<br/>".join(
-                            '<font color="{color}">{prop}: {dt}</font>'.format(
+                            '<font point-size="{fontsize}" color="{color}">{prop}: {dt}</font>'.format(
                                 color="darkgreen" if predicate in self.shapes[class_] else "black",
+                                fontsize=round(node_font_size * 2 / 3),
                                 prop=prop, dt=dt) for predicate, prop, dt in class_data['data'].keys()))
                 node = pydot.Node(name='"' + class_ + '"',
                                   margin="0",
@@ -609,13 +755,29 @@ class OntoGraf:
                 node = pydot.Node(name='"' + class_ + '"',
                                   label=class_data['label'] if class_data['label'] else self.strip_uri(class_),
                                   style='filled',
+                                  fontsize=node_font_size,
+                                  penwidth=node_line_width,
+                                  color="black",
                                   fillcolor="darkgreen" if class_ in self.shapes else "white",
                                   fontcolor="white" if class_ in self.shapes else "black")
 
             self.graf.add_node(node)
 
+            by_predicate = defaultdict(set)
+            if self.concentrate_links != 0:
+                for link in class_data['links']:
+                    predicate, _, target = link
+                    # Don't concentrate self-links
+                    if class_ != target:
+                        by_predicate[predicate].add(link)
+
+            compacted_links = set(predicate for predicate, links in by_predicate.items()
+                                  if len(links) >= self.concentrate_links)
+
             for link, num in class_data['links'].items():
                 predicate, predicate_str, target = link
+                if predicate in compacted_links and target != class_:
+                    continue
                 edge = pydot.Edge(class_, target,
                                   label=predicate_str,
                                   penwidth=self.line_width(num, max_common),
@@ -623,6 +785,33 @@ class OntoGraf:
                                   arrowhead=self.arrowhead)
                 self.graf.add_edge(edge)
 
+            for predicate in compacted_links:
+                links = by_predicate[predicate]
+                shared_node_id = class_ + '_' + predicate
+                shared_node = pydot.Node(name='"' + shared_node_id + '"', shape='point', color="black")
+                self.graf.add_node(shared_node)
+                total_count = sum(class_data['links'][link] for link in links)
+                edge_color = self.shacl_color if predicate in self.shapes[class_] else self.arrow_color
+                predicate_label = next(l for l in links)[1]
+                self.graf.add_edge(
+                    pydot.Edge(class_, shared_node_id,
+                               label=predicate_label,
+                               penwidth=self.line_width(total_count, max_common), color=edge_color))
+                for link in links:
+                    _, _, target = link
+                    edge = pydot.Edge(shared_node_id, target,
+                                      penwidth=self.line_width(class_data['links'][link], max_common),
+                                      color=edge_color,
+                                      arrowhead=self.arrowhead)
+                    self.graf.add_edge(edge)
+
+            for super_class in class_data['supers']:
+                edge = pydot.Edge(class_, super_class,
+                                  penwidth=1,
+                                  color=self.super_color,
+                                  arrowhead='normal')
+                self.graf.add_edge(edge)
+
         self.graf.write(self.outdot)
         if not self.no_image:
             self.graf.write_png(self.outpng)
